from __future__ import annotations


def main() -> None:
    print(
        "\n".join(
            [
                "invertmeeg (invert) — M/EEG inverse solvers",
                "",
                "License: GPLv3 (open source).",
                "If you want to use/ship this in a proprietary product without GPL obligations,",
                "a separate commercial license is available.",
                "Contact: lukas.hecker.job@gmail.com",
                "",
                "Docs: https://lukethehecker.github.io/invertmeeg/",
            ]
        )
    )


if __name__ == "__main__":
    main()
